<?php

namespace App\Http\Controllers;

use App\DataTables\UserShiftDataTable;
use App\Http\Requests;
use App\Http\Requests\CreateUserShiftRequest;
use App\Http\Requests\UpdateUserShiftRequest;
use App\Models\CashRegister;
use App\Models\Coin;
use App\Models\UserShift;
use App\Repositories\UserShiftRepository;
use Carbon\Carbon;
use Flash;
use App\Http\Controllers\AppBaseController;
use Illuminate\Support\Facades\Auth;
use Illuminate\Support\Facades\DB;
use Mockery\Exception;
use Response;

class UserShiftController extends AppBaseController
{
    /** @var  UserShiftRepository */
    private $userShiftRepository;

    public function __construct(UserShiftRepository $userShiftRepo)
    {
        $this->userShiftRepository = $userShiftRepo;
    }

    /**
     * Display a listing of the UserShift.
     *
     * @param UserShiftDataTable $userShiftDataTable
     * @return Response
     */
    public function index(UserShiftDataTable $userShiftDataTable)
    {
        return $userShiftDataTable->render('user_shifts.index');
    }

    /**
     * Show the form for creating a new UserShift.
     *
     * @return Response
     */
    public function create()
    {
        return view('user_shifts.create');
    }

    /**
     * Store a newly created UserShift in storage.
     *
     * @param CreateUserShiftRequest $request
     *
     * @return Response
     */
    public function store(CreateUserShiftRequest $request)
    {
        $input = $request->all();

        $dataShiftUsed = CashRegister::select("id")->whereHas("userShifts",function ($q) use($input){
                $q->where("cash_register_id","=",$input["cash_register_id"])
                    ->where(DB::raw("DATE_FORMAT(start_shift_date, '%Y%m%d')"),"=",Carbon::now()->format("Ymd"))
                    ->where("final_shift_date", "=", null)
                    ->orderBy("start_shift_date", "desc");
            })->get()->toArray();
        if(empty($dataShiftUsed)) {
            try {
                $params = array("user_id" => Auth::user()->id,
                    "cash_register_id" => $request["cash_register_id"],
                    "start_shift_date" => Carbon::now(),
                    "final_shift_date" => null,
                    "init_amount" => $request["init_amount"],
                    "note" => $request["note"]);
                $userShift = $this->userShiftRepository->create($params);

                return response()->json(array("success" => 1, "userShift" => $userShift),200);
            }catch(Exception $e){
                return response()->json(array("msg"=>"Error al crear Shift"),500);
            }
        }else{
            return response()->json(array("msg"=>"Ya hay un turno abierto en esa caja"),500);
        }
    }

    /**
     * Display the specified UserShift.
     *
     * @param  int $id
     *
     * @return Response
     */
    public function show($id)
    {
        $userShift = $this->userShiftRepository->findWithoutFail($id);

        if (empty($userShift)) {
            Flash::error('User Shift not found');

            return redirect(route('userShifts.index'));
        }

        return view('user_shifts.show')->with('userShift', $userShift);
    }

    /**
     * Show the form for editing the specified UserShift.
     *
     * @param  int $id
     *
     * @return Response
     */
    public function edit($id)
    {
        $userShift = $this->userShiftRepository->findWithoutFail($id);

        if (empty($userShift)) {
            Flash::error('User Shift not found');

            return redirect(route('userShifts.index'));
        }

        return view('user_shifts.edit')->with('userShift', $userShift);
    }

    /**
     * Update the specified UserShift in storage.
     *
     * @param  int              $id
     * @param UpdateUserShiftRequest $request
     *
     * @return Response
     */
    public function update($id, UpdateUserShiftRequest $request)
    {
        $userShift = $this->userShiftRepository->findWithoutFail($id);

        if (empty($userShift)) {
            Flash::error('User Shift not found');

            return redirect(route('userShifts.index'));
        }

        $userShift = $this->userShiftRepository->update(array("final_shift_date"=>Carbon::now()->format("Y-m-d H:m:s")), $id);

        Flash::success('User Shift updated successfully.');

        return redirect(route('userShifts.index'));
    }

    /**
     * Remove the specified UserShift from storage.
     *
     * @param  int $id
     *
     * @return Response
     */
    public function destroy($id)
    {
        $userShift = $this->userShiftRepository->findWithoutFail($id);

        if (empty($userShift)) {
            Flash::error('User Shift not found');

            return redirect(route('userShifts.index'));
        }

        $this->userShiftRepository->delete($id);

        Flash::success('User Shift deleted successfully.');

        return redirect(route('userShifts.index'));
    }


    /**
     * Get if exist a ShiftOpen.
     *
     *
     *
     * @return JSON
     */
    public function checkShift()
    {
        $dataShift = UserShift::where("user_id", "=", Auth::user()->id)
            ->orderBy("start_shift_date", "desc")
            ->where("final_shift_date", "=", null)
            ->first();

        if(empty($dataShift)) {
            return response()->json(array("msg"=>"Se requiere abrir un turno para hacer una salida de efectivo"),500);
        }

        return response()->json(array("success"=>1,"message"=>"Ok",
            "data"=>$dataShift),200);
    }


    /**
     * Get CashRegisterDisponible.
     *
     *
     *
     * @return JSON
     */
    public function getCRDisponible()
    {
        

        $dataShiftUsed = CashRegister::select("id")->whereHas("userShifts",function ($q){
                $q->where(DB::raw("DATE_FORMAT(start_shift_date, '%Y%m%d')"),"=",Carbon::now()->format("Ymd"))
                ->where("final_shift_date", "=", null)
                ->orderBy("start_shift_date", "desc");
            })
            ->get();

        $dataShift = CashRegister::whereNotIn('id', $dataShiftUsed)->get();

        return response()->json(array("success"=>1,"message"=>"Ok",
            "data"=>$dataShift),200);
    }

    /**
     * Get CashRegister Cumulative amount.
     *
     *
     *
     * @return JSON
     */
    public function getCRCumulative($cash_register_id)
    {
        $dataShift = CashRegister::whereHas("userShifts",function ($q) use($cash_register_id){
            $q->where(DB::raw("DATE_FORMAT(start_shift_date, '%Y%m%d')"),"=",Carbon::now()->format("Ymd"))
                ->where("final_shift_date", "<>", null)
                ->with("userShifts.tickets")->with("userShifts.tickets.ticketItems")->with("userShifts.tickets.ticketItems.car")->with("userShifts.tickets.ticketItems.car.owner")
                ->with("userShifts.cashOutbound")->with("userShifts.cashOutbound.car")->with("userShifts.cashOutbound.car.owner")
                ->with("userShifts.pay")->with("userShifts.pay.ticket")->with("userShifts.pay.ticket.ticketItems")->with("userShifts.pay.ticket.ticketItems.car")->with("userShifts.pay.ticket.ticketItems.car.owner");
            })
            ->with("userShifts")
            ->where("id","=",$cash_register_id)
            ->first();


        if(isset($dataShift["user_shifts"]["tickets"])) {
            $total_amount_ticket = 0;
            foreach ($dataShift["tickets"] as $ticket) {
                if ($ticket["amount_mx"] > 0) {
                    $total_amount_ticket += floatval($ticket["amoutn_mx"]);
                } else {
                    $total_amount_ticket += floatval($ticket["amount"]);
                }

            }
            $dataShift["total_amount_ticket"] = $total_amount_ticket;
        }else{
            $dataShift["total_amount_ticket"] = 0;
        }

        if(isset($dataShift["user_shifts"]["cashOutbound"])) {
            $total_amount_outbound = 0;
            foreach ($dataShift["cashOutbound"] as $outbound) {
                $total_amount_outbound += floatval($outbound["amount"]);
            }
            $dataShift["total_amount_outbound"] = $total_amount_outbound;
        }else{
            $dataShift["total_amount_outbound"] = 0;
        }

        if(isset($dataShift["user_shifts"]["pay"])) {
            $total_amount_pay = 0;
            foreach ($dataShift["pay"] as $pay) {
                $total_amount_pay += floatval($pay["ticket"]["amount"]);
            }
            $dataShift["total_amount_pay"] = $total_amount_pay;
        }else{
            $dataShift["total_amount_pay"] = 0;
        }

        if(isset($dataShift["user_shifts"])) {
            $dataShift["amount_ini"] = $dataShift["user_shifts"][0]["init_amount"];
        }else{
            $dataShift["amount_ini"] = 0;
        }

        return response()->json(array("success"=>1,"message"=>"Ok",
            "data"=>$dataShift),200);
    }

    /**
     * Get UserShift, Tickets, and totals.
     *
     *
     *
     * @return JSON
     */
    public function getClose()
    {
        $dataShift = UserShift::with("tickets")->with("tickets.ticketItems")->with("tickets.ticketItems.car")->with("tickets.ticketItems.car.owner")
            ->with("tickets.ticketMethodPayment")->with("tickets.ticketMethodPayment.exchangeRate")->with("tickets.ticketMethodPayment.exchangeRate.coin")
            ->with("cashOutbound")->with("cashOutbound.car")->with("cashOutbound.car.owner")
            ->with("pay")->with("pay.ticket")->with("pay.ticket.ticketItems")->with("pay.ticket.ticketItems.car")->with("pay.ticket.ticketItems.car.owner")
            ->with("user")
            ->where("user_id","=",Auth::user()->id)
            ->orderBy("start_shift_date","desc")
            ->where("final_shift_date","=",null)

            ->first();

        $coinData = Coin::pluck("id","id")->toArray();

        $coinData["userShift"]["id"] = $dataShift["id"];
        $coinData["userShift"]["user_id"] = $dataShift["user_id"];
        $coinData["userShift"]["cash_register_id"] = $dataShift["cash_register_id"];
        $coinData["userShift"]["start_shift_date"] = $dataShift["start_shift_date"];
        $coinData["userShift"]["final_shift_date"] = $dataShift["final_shift_date"];
        $coinData["userShift"]["init_amount"] = $dataShift["init_amount"];



        if(isset($dataShift["tickets"])) {
            $total_amount_ticket = 0;
            foreach ($dataShift["tickets"] as $t => $ticket) {
                foreach($ticket["ticketMethodPayment"] as $mp => $methodPayment) {

                    if(!isset($coinData[$methodPayment["exchangeRate"]["coin_id"]]["tickets"])){
                        $coinData[$methodPayment["exchangeRate"]["coin_id"]] = [];
                        $coinData[$methodPayment["exchangeRate"]["coin_id"]]["tickets"]=[];
                    }

                    $coinData[$methodPayment["exchangeRate"]["coin_id"]]["tickets"][$t] = $ticket;


                    if($methodPayment["methodPayment"]["type"] == "efectivo") {
                        if (floatval($methodPayment["exchangeRate"]["exchange_rate"]) == 1) {
                            if (!isset($coinData[$methodPayment["exchangeRate"]["coin_id"]]["total_amount_ticket"])) {
                                $coinData[$methodPayment["exchangeRate"]["coin_id"]]["total_amount_ticket"] = floatval($ticket["amount_mx"]);
                            } else {
                                $coinData[$methodPayment["exchangeRate"]["coin_id"]]["total_amount_ticket"] += floatval($ticket["amount_mx"]);
                            }

                        } else {
                            if (!isset($coinData[$methodPayment["exchangeRate"]["coin_id"]]["total_amount_ticket"])) {
                                $coinData[$methodPayment["exchangeRate"]["coin_id"]]["total_amount_ticket"] = floatval($ticket["amount"]);
                            } else {
                                $coinData[$methodPayment["exchangeRate"]["coin_id"]]["total_amount_ticket"] += floatval($ticket["amount"]);
                            }
                        }
                    }
                }
            }
            //$dataShift["total_amount_ticket"] = $total_amount_ticket;
        }

        if(!is_array($coinData[2])) {
            $coinData[2] = [];
            $coinData[2]["total_amount_ticket"] = 0;
        }

        $total_amount_outbound = 0;
        if (isset($dataShift["cashOutbound"])) {
            foreach ($dataShift["cashOutbound"] as $outbound) {

                if (!isset($coinData[2]["total_amount_outbound"])) {
                    $total_amount_outbound = floatval($outbound["amount"]);
                } else {
                    $total_amount_outbound += floatval($outbound["amount"]);
                }
            }
        }
        $coinData[2]["total_amount_outbound"] = $total_amount_outbound;

        $total_amount_pay = 0;
        if (isset($dataShift["pay"])) {
            foreach ($dataShift["pay"] as $pay) {
                $total_amount_pay += floatval($pay["ticket"]["amount"]);
            }
        }
        $coinData[2]["total_amount_pay"] = $total_amount_pay;

        if (empty($dataShift)) {
            return response()->json(array("success"=>0,"msg"=>"No se encontro un Turno Abierto"));
        }

        return response()->json(array("success"=>1,"msg"=>"Correcto",
            "data"=>$coinData));
    }


    /**
     * Get UserShift, Tickets, and totals.
     *
     *
     *
     * @return JSON
     */
    public function viewClose($id)
    {
        $dataShift = UserShift::with("tickets")->with("tickets.ticketItems")->with("tickets.ticketItems.car")->with("tickets.ticketItems.car.owner")
            ->with("tickets.ticketMethodPayment")->with("tickets.ticketMethodPayment.exchangeRate")->with("tickets.ticketMethodPayment.exchangeRate.coin")
            ->with("cashOutbound")->with("cashOutbound.car")->with("cashOutbound.car.owner")
            ->with("pay")->with("pay.ticket")->with("pay.ticket.ticketItems")->with("pay.ticket.ticketItems.car")->with("pay.ticket.ticketItems.car.owner")
            ->with("user")
            ->orderBy("start_shift_date","desc")
            ->where("user_shift.id","=",$id)
            ->first();

        $coinData = Coin::pluck("id","id")->toArray();

        $coinData["userShift"]["id"] = $dataShift["id"];
        $coinData["userShift"]["user_id"] = $dataShift["user_id"];
        $coinData["userShift"]["cash_register_id"] = $dataShift["cash_register_id"];
        $coinData["userShift"]["start_shift_date"] = $dataShift["start_shift_date"];
        $coinData["userShift"]["final_shift_date"] = $dataShift["final_shift_date"];
        $coinData["userShift"]["init_amount"] = $dataShift["init_amount"];



        if(isset($dataShift["tickets"])) {
            $total_amount_ticket = 0;
            foreach ($dataShift["tickets"] as $t => $ticket) {
                foreach($ticket["ticketMethodPayment"] as $mp => $methodPayment) {

                    if(!isset($coinData[$methodPayment["exchangeRate"]["coin_id"]]["tickets"])){
                        $coinData[$methodPayment["exchangeRate"]["coin_id"]] = [];
                        $coinData[$methodPayment["exchangeRate"]["coin_id"]]["tickets"]=[];
                    }

                    $coinData[$methodPayment["exchangeRate"]["coin_id"]]["tickets"][$t] = $ticket;


                    if (floatval($methodPayment["exchangeRate"]["exchange_rate"]) == 1) {
                        if(!isset($coinData[$methodPayment["exchangeRate"]["coin_id"]]["total_amount_ticket"])) {
                            $coinData[$methodPayment["exchangeRate"]["coin_id"]]["total_amount_ticket"] = floatval($ticket["amount_mx"]);
                        }else{
                            $coinData[$methodPayment["exchangeRate"]["coin_id"]]["total_amount_ticket"] += floatval($ticket["amount_mx"]);
                        }

                    }else{
                        if(!isset($coinData[$methodPayment["exchangeRate"]["coin_id"]]["total_amount_ticket"])) {
                            $coinData[$methodPayment["exchangeRate"]["coin_id"]]["total_amount_ticket"] = floatval($ticket["amount"]);
                        }else{
                            $coinData[$methodPayment["exchangeRate"]["coin_id"]]["total_amount_ticket"] += floatval($ticket["amount"]);
                        }
                    }
                }
            }
            //$dataShift["total_amount_ticket"] = $total_amount_ticket;
        }

        if(!is_array($coinData[2])) {
            $coinData[2] = [];
            $coinData[2]["total_amount_ticket"] = 0;
        }

        $total_amount_outbound = 0;
        if (isset($dataShift["cashOutbound"])) {
            foreach ($dataShift["cashOutbound"] as $outbound) {

                if (!isset($coinData[2]["total_amount_outbound"])) {
                    $total_amount_outbound = floatval($outbound["amount"]);
                } else {
                    $total_amount_outbound += floatval($outbound["amount"]);
                }
            }
        }
        $coinData[2]["total_amount_outbound"] = $total_amount_outbound;

        $total_amount_pay = 0;
        if (isset($dataShift["pay"])) {
            foreach ($dataShift["pay"] as $pay) {
                $total_amount_pay += floatval($pay["ticket"]["amount"]);
            }
        }
        $coinData[2]["total_amount_pay"] = $total_amount_pay;



        if (empty($dataShift)) {
            return response()->json(array("success"=>0,"msg"=>"No se encontro un Turno Abierto"));
        }

        return response()->json(array("success"=>1,"msg"=>"Correcto",
            "data"=>$coinData));
    }



}
